
/***********************************************************************************/
/*                           The One-Way Layout                                    */
/***********************************************************************************/

%macro OWL(DATA = _LAST_,
           VAR =,
           GROUP =,
           ALPHA_C = 0.05,
           ALPHA_P = 0.05,
           EXACT = YES,
           N_SIM = 1000,
           DATA_PT = _NOPT_,
           VAR_PT =,
           GROUP_PT =);

PROC IML;

/***********************************************************************************/
/*                        Read the SAS Data-Set                                    */
/***********************************************************************************/
sort &DATA by &GROUP;
use &DATA;
   read all var{&GROUP} into grp;
   read all var{&VAR} into var;
close &DATA;

/***********************************************************************************/
/*                Replace the Class-Levels by 1, ..., n                            */
/***********************************************************************************/
names = union(grp);
help = j(nrow(grp), 1);
do i = 1 to ncol(names);
   help[loc(grp = names[i])] = i;
end;
grp = help;

/***********************************************************************************/
/*                      Sample-Sizes, Ranks, ...                                   */
/***********************************************************************************/
n = nrow(grp);
a = max(grp);
m = max(var) + 1;
rij = ranktie(var);
ni = 1 : a;
riq = 1 : a;
riij = 1 : n;
rkji = repeat(var, 1, a);
sigmaq = 1 : a;
do i = 1 to a;
   ni[i] = ncol(loc(grp = i));
   riq[i] = rij[loc(grp = i)][:];
   sigmaq[i] = ssq(rij[loc(grp = i)] - ranktie(var[loc(grp = i)]) - riq[i]
             + (ni[i] + 1) / 2) / (ni[i] - 1);
   rkji[loc(grp = i), i] = m;
   rkji[, i] = ranktie(rkji[, i]);
   if i = 1 then
      mat = j(1, ni[1], 1 / ni[1]);
   else
      mat = block(mat, j(1, ni[i], 1 / ni[i]));
end;

/***********************************************************************************/
/*                       Kruskal-Wallis-Test                                       */
/***********************************************************************************/
s1q = ssq(rij - (n + 1) / 2) / (n -1);
cmp = ni * ((riq - (n + 1) / 2) ## 2)`;
qn = cmp / s1q;
pn = 1 -probchi(qn, a - 1);

/***********************************************************************************/
/*                    F-Test Rank Statistic                                        */
/***********************************************************************************/
s2q = ssq(rij - riq[grp]) / (n - a);
qnf = cmp / s2q;
pnfl = 1 - probchi(qnf, a - 1);

/***********************************************************************************/
/*                         Simulation                                              */
/***********************************************************************************/
psim = 0;
rm = (n + 1) / 2;
do i = 1 to &N_SIM;
   psim = psim + (cmp <= ni * (mat * rij[rank(uniform(j(n, 1, 0)))] - rm) ## 2);
end;
psim = psim / &N_SIM;

/***********************************************************************************/
/*                       F-Approximation                                           */
/***********************************************************************************/
pnfs = 1 - probf(qnf / (a - 1), a - 1, n - a - 1);

/***********************************************************************************/
/*                    Confidence-Intervals                                         */
/***********************************************************************************/
pd = (riq - 1 / 2) / n;

zah = probit(1 - &ALPHA_C / 2);
tauq = j(a, a, 0);
sq = 1 : a;
pl = 1 : a;
pu = 1 : a;
do i = 1 to a;
   do k = 1 to a;
     if k ^= i then
        tauq[k, i] = ssq(rij[loc(grp = k)] - rkji[loc(grp = k), i] - riq[k]
                   + rkji[loc(grp = k), i][:]) / (ni[k] - 1);
   end;
   sq[i] = sigmaq[i] / (n * ni[i]) + ni * tauq[, i] / (n * ni[i] ** 2);
   pl[i] = pd[i] - zah * sqrt(sq[i] / n);
   pu[i] = pd[i] + zah * sqrt(sq[i] / n);
end;

/***********************************************************************************/
/*                      Multiple Comparisons                                       */
/***********************************************************************************/
j = 1;
mcomp = j(a * (a - 1) / 2, 12, 0);
do i = 1 to a;
   do is = i + 1 to a;
      mcomp[j, 1] = i;
      mcomp[j, 2] = is;
      mcomp[j, 3] = (riq[i] -riq[is]) / sqrt(((ssq(rij[loc(grp = i)] - riq[i])
                  + ssq(rij[loc(grp = is)] - riq[is])) / (ni[i] + ni[is] - 2))
                  * (1 / ni[i] + 1 / ni[is]));
      mcomp[j, 4] = 2 * (1 - probnorm(abs(mcomp[j, 3])));
      mcomp[j, 5] = 2 * (1 - probt(abs(mcomp[j, 3]), ni[i] + ni[is] -3));

      /* Exact p-Values for Pairwise Comaprisons */
      %if &EXACT = YES %then %do;
         rsmc = 2 * ni[is] *riq[is];
         rmc = 2 * (rij[loc(grp = i | grp = is)]);
         rmcr = rank(rmc);
         mat = 1;
         dim = 1;
         do k = 1 to ni[i] + ni[is];
            r = rmc[loc(rmcr = k)];
            dim = dim + r;
            mat = (mat || j(k, r, 0) // j(1, dim, 0))
                + (j(1, dim, 0) // (j(k, r, 0) || mat));
         end;
         zv = cusum(mat[ni[is] + 1, ]);
         zv = zv / max (zv);
         pwel = 1 - zv[rsmc];
         pwer = zv[rsmc + 1];
         mcomp[j, 6] = 2 * min(pwel, pwer, 0.5);
      %end;

      j = j + 1;
   end;
end;

/* Holm-Procedure */
mcomp[, 7] = rank(mcomp[, 4]);
mcomp[, 8] = rank(mcomp[, 5]);
mcomp[, 9] = rank(mcomp[, 6]);
phil = 1;
phis = 1;
phie = 1;
do i = 1 to a * (a - 1) / 2;
   posl = loc(mcomp[, 7] = i);
   poss = loc(mcomp[, 8] = i);
   pose = loc(mcomp[, 9] = i);
   phil = mcomp[posl, 4] <= &ALPHA_P / (a * (a - 1) / 2 + 1 - i) & phil;
   phis = mcomp[posl, 5] <= &ALPHA_P / (a * (a - 1) / 2 + 1 - i) & phis;
   phie = mcomp[posl, 6] <= &ALPHA_P / (a * (a - 1) / 2 + 1 - i) & phie;
   mcomp[posl, 10] = phil;
   mcomp[poss, 11] = phis;
   mcomp[pose, 12] = phie;
end;

%if &EXACT ^= YES %then %do;
   mcomp[, 6] = j(a * (a - 1) / 2, 1, .);
   mcomp[, 9] = j(a * (a - 1) / 2, 1, .);
   mcomp[, 12] = j(a * (a - 1) / 2, 1, .);
%end;

/***********************************************************************************/
/*                          Patterned Alternatives                                 */
/***********************************************************************************/
%if &DATA_PT ^= _NOPT_ %then %do;
   sort &DATA_PT by &GROUP_PT;
   use &DATA_PT;
      read all var{&VAR_PT} into wi;
   close &DATA_PT;

   wic = wi - wi[:];
   fa = (ni ## -1) * (wic ## 2);
   slq = fa * ssq(rij - riq[grp]) / (n - a);
   ln = riq * wic / sqrt(slq);
   ppl = 1 - probnorm(ln);
   pps = 1 - probt(ln, n - a - 1);
%end;

/***********************************************************************************/
/*                                  Output                                         */
/***********************************************************************************/
reset linesize = 80;
reset noname;

print "NONPARAMETRIC", "ONE-WAY LAYOUT";
print "Data Information, Estimation, Confidence Intervals (alpha = &ALPHA_C)";
print "Data Set: &DATA";
print "Total Sample Size:" n;
print "                                                            Confidence Intervals",
      ((1 : a)`) [c = "Nr."]
      (names`) [c = "Class Levels"]
      (ni`) [c = "n_i"]
      (riq`) [c = "Rank Means"]
      (pd`) [c = "p_i"]
      (pl`) [c = "p_L"]
      (pu`) [c = "p_U"],, ;
print "Hypothesis Testing: F_1 = ... = F_a";
print "  p-Values",
      (qn // qnf) [c = "Statistic" r = {"Kruskal-Wallis" "F-Test Rank"}]
      (pn // pnfl) [c = "    Chi-Sq. Appr."]
      (. // pnfs) [c = "F-Appr."]
      (psim // .) [c = "Exact"];
print "Number of Simulations for the Exact p-Value: &N_SIM" / ;

print "Pairwise Comparisons (alpha = &ALPHA_P)";
print "  Samples Statistic           p-Values              Decision (Holm-Proc.) ",
      (mcomp[, 1 : 2]) [f = 4.0]
      (mcomp[, 3 : 6]) [c = {"" "Normal" "t-Appr." "Exact"}]
      (mcomp[, 10 : 12]) [f = 1.0 c = {"Normal" "t-Appr." "Exact"}],,;
%if &DATA_PT ^= _NOPT_ %then %do;
   print "Patterned Alternatives";
   print "Pattern:" (wi`) [f = 3.0];
   print "                    p-Values",
         (ln || ppl || pps) [c = {"Statistic" "Normal-Approx." "t-Approx."}];
%end;

QUIT;

%mend OWL;
